from herre.graphical import GraphicalBackend
from herre import get_current_herre

class UserDeniedAssignation(Exception):
    pass



def fill_args_kwargs_graphically(node, args, kwargs, parent=None):

    with GraphicalBackend(parent=parent):
            from arkitekt.ui.assign_widget import AssignWidget
            assignWidget = AssignWidget(node=node,set_args=args, set_kwargs=kwargs)
            if assignWidget.exec_():
                return assignWidget.args_kwargs_tuple
            else:
                raise UserDeniedAssignation("User Rejected an Assignation")











CURRENT_RPC = None
CURRENT_AGENT = None


def get_current_rpc(force_creation=False):
    global CURRENT_RPC
    if CURRENT_RPC is None and force_creation:
        from arkitekt.rpc import RPC
        CURRENT_RPC = RPC()
    return CURRENT_RPC


def set_current_rpc(arkitekt, overwrite=False):
    global CURRENT_RPC
    CURRENT_RPC = arkitekt


def get_current_agent(force_creation=False):
    global CURRENT_AGENT
    if CURRENT_AGENT is None and force_creation:
        from arkitekt.agents import AppAgent
        CURRENT_AGENT = AppAgent()
    return CURRENT_AGENT


def set_current_agent(arkitekt, overwrite=False):
    global CURRENT_AGENT
    CURRENT_AGENT = arkitekt


from herre.wards.variables import QueryVariable
from typing import Any, List, Optional
from pydantic.class_validators import validator
from pydantic.fields import Field
from pydantic.main import BaseModel
from herre.wards.graphql import ParsedQuery, GraphQLWard
from arkitekt.models.base import AsyncModelManager, SyncModelManager, Model, ModelType

class GraphQLSyncModelManager(SyncModelManager):

    def from_query(self, query, **kwargs):
        res_dict =  self.ward._run_sync(query, variables=kwargs)
        return self.modelClass(**res_dict[query.firstchild])


    def get(self, **kwargs) -> ModelType:
        return self.from_query(self.modelClass.Meta.get, **kwargs)

    def create(self, **kwargs) -> ModelType:
        return self.from_query(self.modelClass.Meta.create, **kwargs)


    
class GraphQLAsyncModelManager(AsyncModelManager):


    async def from_query(self, query, **kwargs):
        res_dict =  await self.ward._run_async(query, variables=kwargs)
        return self.modelClass(**res_dict[query.firstchild])
    
    async def get(self, **kwargs) -> ModelType:
        return await self.from_query(self.modelClass.Meta.get, **kwargs)

    async def create(self, **kwargs) -> ModelType:
        return await self.from_query(self.modelClass.Meta.create, **kwargs)


        

# Mixin CLass is not Possible because of BaseModel Metaclass
class GraphQLObject(BaseModel):
    typename: Optional[str] = Field(alias="__typename")

    def __init__(__pydantic_self__, **data: Any) -> None:
        if "__typename" not in data: data["__typename"] = __pydantic_self__.__class__.__name__ 
        super().__init__(**data)

    @validator('typename')
    def typename_matches_class(cls, v):
        if v is None: return None # We are ommiting typechecks if __typename is not explicitly set
        if cls.__name__ == v: return v
        raise ValueError(f"Didn't find correct class {cls.__name__} __typename {v}")

    def dict(self, *args, by_alias=True, **kwargs):
        return super().dict(*args, **{
            **kwargs,
            "by_alias": by_alias,
        })

    


class GraphQLModel(Model, QueryVariable):
    typename: Optional[str] = Field(alias="__typename")


    def __init__(__pydantic_self__, **data: Any) -> None:
        if "__typename" not in data: data["__typename"] = __pydantic_self__.__class__.__name__ 
        super().__init__(**data)


    @validator('typename')
    def typename_matches_class(cls, v):
        if v is None: return None # We are ommiting typechecks if __typename is not explicitly set
        if cls.__name__ == v: return v
        raise ValueError(f"Didn't find correct class {cls.__name__} __typename {v}")

    def dict(self, *args, as_input=False, **kwargs):
        return super().dict(*args, **{
            **kwargs,
            "by_alias": not as_input,
        })


    async def to_variable(self):
        return self.id

    @classmethod
    def get_objectsclass(cls):
        return GraphQLSyncModelManager

    @classmethod
    def get_asyncclass(cls):
        return GraphQLAsyncModelManager

    @property
    def ward(self) -> GraphQLWard:
        return self.__class__.asyncs.ward

    class Meta:
        abstract = True


from enum import Enum
from typing import Optional, Union
from arkitekt.models.graphql import GraphQLModel, GraphQLObject

class ReservePolicy(GraphQLObject):
    """Determinns how the once running
    provision can be reserved

    Args:
        GraphQLModel ([type]): [description]
    """
    pass

class PrivateReservePolicy(ReservePolicy):
    # Only the provisioner can reserve a provision of this template
    pass


class AllReservePolicy(ReservePolicy):
    # All users can reserve a provision of this template
    pass

class GroupReservePolicy(ReservePolicy):
    # Only users of a certain group can reserve this provision
    allowed_groups: Optional[int]    


class ProvisionPolicy(GraphQLObject):
    """Determins who and when a user can
    provision a pod also, if it should die or not
    after

    Args:
        GraphQLModel ([type]): [description]
    """
    pass

class AllProvisionPolicy(ProvisionPolicy):

    pass

class AdminProvisionPolicy(ProvisionPolicy):

    pass


class GroupProvisionPolicy(ProvisionPolicy):
    allowed_groups: Optional[int]  


class BalanceStrategy(GraphQLObject):
    """
    Distinguihhes how reservations to this worker are being handles
    should we try to provision a new worker for every reservation, etc
    
    """

    pass

class OneToOneStrategy(BalanceStrategy):
    """Map every Reservation to its own Provision

    Args:
        BalanceStrategy ([type]): [description]
    """
    allowed_provisions_per_user: int = -1 # indefinite amount of provisions allowed

class ManyToOneStrategy(BalanceStrategy):
    """Share a provision amongs a lot of reservations

    Use Cases:
        - A lot of users can access one resource (e.g. Microscope) without saving
          intermediate state (e.g a communal slice is mounted)

    Args:
        BalanceStrategy ([type]): [description]
    """


class OneToManyStrategy(BalanceStrategy):
    """Map every Reservation to its own Provision

    Args:
        BalanceStrategy ([type]): [description]
    """





class Policy(GraphQLObject):
    reserve: ReservePolicy
    provision: Union[AllProvisionPolicy, GroupProvisionPolicy]
    strategy: Union[OneToOneStrategy, ManyToOneStrategy, OneToManyStrategy]



EACH_USER_ONE_PROVISION = Policy(provision=AllProvisionPolicy(), reserve=AllReservePolicy(), strategy=OneToOneStrategy(allowed_provisions_per_user=1))
EVERY_USER_ON_PROVISION = Policy(provision=AllProvisionPolicy(), reserve=AllReservePolicy(), strategy=ManyToOneStrategy(allowed_provisions_per_user=1))
ONE_INSTANCE_PROVIDABLE_BY_ADMIN_SHARED_BY_EVERYONE = Policy(provision=AdminProvisionPolicy(), reserve=AllReservePolicy(), strategy=ManyToOneStrategy(allowed_provisions_per_user=1))
MANY_INSTANCES_PROVIDABLE_BY_ADMIN_SHARED_BY_GROUP = Policy(provision=AdminProvisionPolicy(), reserve=AllReservePolicy(), strategy=ManyToOneStrategy(allowed_provisions_per_user=-1))



from abc import abstractmethod
from arkitekt.ui.qtwidgets.base import UIPortMixin

from pydantic.main import BaseModel
from arkitekt.models.graphql import GraphQLModel
from arkitekt.packers.registry import get_packer_registry
from re import search
from typing import Any, Generic, Type, TypeVar
from arkitekt.schema.ports import ArgPort, KwargPort, ListArgPort, ListKwargPort, Port, StructureArgPort, StructureKwargPort
from arkitekt.schema.widgets import SearchWidget
from qtpy import QtWidgets
from arkitekt.schema.node import Node
from itertools import zip_longest
from typing import Generic
from arkitekt.ui.registry import get_widget_registry




def build_qtwidget_for_argport(port: ArgPort):
    qtwidgetClass = get_widget_registry().get_widget_class(port.typename, port.widget.typename)
    return qtwidgetClass(port=port)

def build_qtwidget_for_kwargport(port: KwargPort):
    qtwidgetClass = get_widget_registry().get_widget_class(port.typename, port.widget.typename)
    return qtwidgetClass(port=port)



class AssignWidget(QtWidgets.QDialog):

    def __init__(self, *args, node: Node = None, set_args = [], set_kwargs={}, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        assert node is not None, "Provide the node Parameter to use the assign widget"
        self.node = node
        self.set_args = set_args
        self.set_kwargs = set_kwargs
        self.populated_args = []
        self.populated_kwargs = {}
        self.args_kwargs_tuple = ([],{})


        assert len(self.set_args) < len(self.node.args), "Was provided more Args than node accepts"


        self.build_ui()

    def build_args_widget(self):
        self.argsWidget = QtWidgets.QGroupBox("Args")
        self.argsLayout = QtWidgets.QFormLayout()

        for arg, port in zip_longest(self.set_args, self.node.args):
            if arg: 
                self.populated_args.append(arg)
            else:
                widget = build_qtwidget_for_argport(port)
                self.argsLayout.addRow(QtWidgets.QLabel(parent=self, text=port.label or port.key), widget)
                self.populated_args.append(widget)

        self.argsWidget.setLayout(self.argsLayout)
        return self.argsWidget

    def build_kwargs_widget(self):
        self.kwargsWidget = QtWidgets.QGroupBox("Kwargs")
        self.kwargsLayout = QtWidgets.QFormLayout()

        for port in self.node.kwargs:
            if port.key in self.set_kwargs: 
                self.populated_kwargs[port.key] = self.set_kwargs[port.key]
            else:
                widget = build_qtwidget_for_kwargport(port)
                self.kwargsLayout.addRow(QtWidgets.QLabel(parent=self, text=port.label or port.key), widget)
                self.populated_kwargs[port.key] = widget

        self.kwargsWidget.setLayout(self.kwargsLayout)
        return self.kwargsWidget
            
                # Arg is not Set so we have to build a widget


    def build_ui(self):

        self.setWindowTitle(f"Assigning to {self.node.name} : {self.node.package}/{self.node.interface}")

        self.layout = QtWidgets.QVBoxLayout()

        self.node_description =  QtWidgets.QLabel(self.node.description)

        self.argsWidget = self.build_args_widget()
        self.kwargsWidget = self.build_kwargs_widget()


        self.buttonBox = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel)
        self.buttonBox.accepted.connect(self.on_accept)
        self.buttonBox.rejected.connect(self.on_reject)

        self.layout.addWidget(self.node_description)
        self.layout.addWidget(self.argsWidget)
        self.layout.addWidget(self.kwargsWidget)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)

    def on_accept(self):
        done_args = []

        for arg in self.populated_args:
            if isinstance(arg, UIPortMixin):
                done_args.append(arg.get_current_value())
            else:
                done_args.append(arg)
                    

        done_kwargs = {}

        for key, kwarg in self.populated_kwargs.items():
            if isinstance(kwarg, UIPortMixin):
                done_kwargs[key] = kwarg.get_current_value()
            else:
                done_kwargs[key] = kwarg


        self.args_kwargs_tuple = (done_args, done_kwargs)
        self.accept()
        self.close()


    def on_reject(self):
        self.args_kwargs_tuple = ([], {})
        self.accept()
        self.close()









from ..messages.base import MessageDataModel, MessageMetaModel, MessageMetaExtensionsModel, MessageModel

from typing import Optional
from ..messages.types import EXCEPTION

class ProtocolException(Exception):
    pass

class ExceptionMetaModel(MessageMetaModel):
    type: str = EXCEPTION
    reference: str

class ExceptionDataModel(MessageDataModel):
    message: Optional[str]
    type: str = "Exception"


class ExceptionMessage(MessageModel):
    data: ExceptionDataModel
    meta: ExceptionMetaModel = {"type": EXCEPTION}


    @classmethod
    def fromException(cls, e: Exception, reference):
        return cls(data={"message": str(e), "type": e.__class__.__name__}, meta={"reference": reference})


    def toException(self):
        return ProtocolException(f'{self.data.type}: {self.data.message}')


from arkitekt.schema.negotiation import Transcript
from herre.config.model import BaseConfig
from herre.wards.graphql import ParsedQuery, GraphQLWard
from herre.auth import HerreClient


class ArkitektConfig(BaseConfig):
    _group = "arkitekt"
    host: str
    port: int
    secure: bool


class ArkitektWard(GraphQLWard):

    class Meta:
        key = "arkitekt"

    def __init__(self, herre: HerreClient) -> None:
        self.config = ArkitektConfig.from_file(herre.config_path)
        self.transcript: Transcript = None
        super().__init__(herre, f"http://{self.config.host}:{self.config.port}/graphql")


    async def negotiate(self):
        transcript_query = await self.run(ParsedQuery("""mutation Negotiate {
            negotiate {
                postman {
                    type
                    kwargs
                }
            } 
        }"""))
        return Transcript(**transcript_query["negotiate"])



from qtpy import QtWidgets
from abc import abstractmethod

class UIPortMixin:

    def __init__(self, *args, port = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.port = port


    @abstractmethod
    def get_current_value(self):
        """Gets the current value of the widget or returns None if
        no user input was set, or default

        Raises:
            NoValueSetError: An error that this widgets has no current value set
        """
        raise NotImplementedError("Please overwrite this in your Widget class")

class ErrorWidgetException(Exception):
    pass


class ErrorWidget(QtWidgets.QWidget, UIPortMixin):
    
    def __init__(self, *args, port, **kwargs) -> None:
        super().__init__(*args, port=port, **kwargs)
        self.build_ui()


    def build_ui(self):
        self.layout = QtWidgets.QVBoxLayout()
        self.searchBox = QtWidgets.QLineEdit(f"Error Widget for port {self.port.key} {self.port.typename} {self.port.widget.typename if self.port.widget else 'No widget provided'}")
        self.layout.addWidget(self.searchBox)
        self.setLayout(self.layout)

    def get_current_value(self):
        raise ErrorWidgetException("There was not Widget installed for this type")


import contextvars
from arkitekt.messages.postman.assign.assign_critical import AssignCriticalMessage
from arkitekt.messages.postman.assign.assign_cancelled import AssignCancelledMessage
from arkitekt.messages.postman.assign.assign_return import AssignReturnMessage
from arkitekt.messages.postman.assign.assign_yield import AssignYieldsMessage
from arkitekt.messages.postman.assign.assign_done import AssignDoneMessage

from arkitekt.messages.postman.assign.bounced_forwarded_assign import BouncedForwardedAssignMessage
from arkitekt.actors.base import Actor
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
import asyncio
from arkitekt.packers.utils import shrink_inputs, shrink_outputs, expand_inputs, expand_outputs
import logging

logger = logging.getLogger(__name__)


class FunctionalActor(Actor):
    pass


assign_message = contextvars.ContextVar("assign_message")



class FunctionalFuncActor(FunctionalActor):
    
    async def progress(self, value, percentage):
        await self._progress(value, percentage)

    async def assign(self, *args, **kwargs):
        raise NotImplementedError("Please provide a func or overwrite the assign method!")


    async def on_assign(self, message: BouncedForwardedAssignMessage):
        try:
            logger.info("Assigning Number two")
            args, kwargs = await expand_inputs(self.template.node, message.data.args, message.data.kwargs) if self.expand_inputs else (message.data.args, message.data.kwargs)
            
            assign_message.set(message)
            returns = await self.assign(*args, **kwargs)
            assign_message.set(None)

            await self.transport.forward(AssignReturnMessage(data={
                "returns": await shrink_outputs(self.template.node, returns) if self.shrink_outputs else returns
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))


        except asyncio.CancelledError as e:

            await self.transport.forward(AssignCancelledMessage(data={
                "canceller": str(e)
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))



        except Exception as e:
            self.console.print_exception()
            await self.transport.forward(AssignCriticalMessage(data={
                "type": e.__class__.__name__,
                "message": str(e)
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))




class FunctionalGenActor(FunctionalActor):

    async def assign(self, *args, **kwargs):
        raise NotImplementedError("Please provide a func or overwrite the assign method!")

    async def progress(self, value, percentage):
        await self._progress(value, percentage)

    async def on_assign(self, message: BouncedForwardedAssignMessage):
        try:
            args, kwargs = await expand_inputs(self.template.node, message.data.args, message.data.kwargs) if self.expand_inputs else (message.data.args, message.data.kwargs)
            
            assign_message.set(message)

            async for returns in self.assign(*args, **kwargs):
                await self.transport.forward(AssignYieldsMessage(data={
                    "returns": await shrink_outputs(self.template.node, returns) if self.shrink_outputs else returns
                }, meta = {
                    "reference": message.meta.reference,
                    "extensions": message.meta.extensions
                }))


            assign_message.set(None)

            await self.transport.forward(AssignDoneMessage(data={
                "returns": await shrink_outputs(self.template.node, returns) if self.shrink_outputs else returns
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))


        except asyncio.CancelledError as e:

            await self.transport.forward(AssignCancelledMessage(data={
                "canceller": str(e)
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))



        except Exception as e:
            self.console.print_exception()
            await self.transport.forward(AssignCriticalMessage(data={
                "type": e.__class__.__name__,
                "message": str(e)
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))



class FunctionalThreadedFuncActor(FunctionalActor):
    nworkers = 5

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.threadpool = ThreadPoolExecutor(self.nworkers)

    def assign(self, *args, **kwargs):
        raise NotImplementedError("")




class FunctionalThreadedGenActor(FunctionalActor):
    nworkers = 5

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.threadpool = ThreadPoolExecutor(self.nworkers)

    def assign(self, *args, **kwargs):
        raise NotImplementedError("")  


from typing import Callable, Coroutine, Type, TypeVar
import inspect

class PackerError(Exception):
    pass

class UnpackableError(PackerError):
    pass

class NoStructureRegisteredError(PackerError):
    pass

class StructureOverwriteError(PackerError):
    pass


class PackerRegistry:

    def __init__(self) -> None:
        self.identifierStructureMap = {}

    def register_structure(self, structure, overwrite=False):

        try:
            assert inspect.ismethod(structure.get_identifier) and structure.get_identifier.__self__ is structure, "Get Identifier must be either provided for be a classmethod"
            assert inspect.ismethod(structure.expand) and structure.expand.__self__ is structure,"Expand Function must be an async classmethod"
            assert inspect.iscoroutinefunction(structure.expand), "Expand Function must be an (classmethod) async function"
            assert inspect.iscoroutinefunction(structure.shrink), "Shrink Function must be an async function"

        except Exception as e:
            raise UnpackableError(str(e)) from e

        identifier = structure.get_identifier()
        if identifier in self.identifierStructureMap and not overwrite:
            raise StructureOverwriteError(f"Another Structure was registerd for identifier '{identifier}'. You can choose to overwrite this Structure by setting overwrite=True")
        self.identifierStructureMap[identifier] = structure


    def get_structure(self, identifier):
        try:
            return self.identifierStructureMap[identifier]
        except KeyError as e:
            raise NoStructureRegisteredError(f"No Structure registered for identifier {identifier}") from e

PACKER_REGISTRY = None

def get_packer_registry():
    global PACKER_REGISTRY
    if not PACKER_REGISTRY:
        PACKER_REGISTRY = PackerRegistry()
    return PACKER_REGISTRY



def register_structure(
    identifier: str = None,
    overwrite: bool = False,
    registry: PackerRegistry=  None,
):
    """Registers a Structure

    Registers a Structure with the a Package Registry. Once registered Arkitekt will use this Structure to expand and shrink Node requests.

    Args:
        identifier (str, optional): A unique identifier to be known be. Defaults to None.
        overwrite (bool, optional): If overwrite is set this Structure will replace exisiting Structures, otherwise we return an error on Reigstration
        registry (PackerRegistry, optional): Which registry to use in order to register this Strucute. Will use the default Registry if not set
    """

    def real_decorator(cls):
        if identifier: cls.get_identifier = classmethod(lambda cls: identifier)
        (registry or get_packer_registry()).register_structure(cls, overwrite=overwrite)
        return cls


    return real_decorator


















from abc import abstractmethod
from arkitekt.ui.sync_funcs import fill_args_kwargs_graphically
from arkitekt.schema.enums import NodeType
from arkitekt.schema.ports import DictArgPort, DictKwargPort, DictReturnPort, IntArgPort, IntKwargPort, IntReturnPort, ListArgPort, ListKwargPort, ListReturnPort, StringArgPort, StringKwargPort, StringReturnPort, StructureArgPort, StructureKwargPort, StructureReturnPort
from arkitekt.graphql.node import NODE_CREATE_QUERY, NODE_GET_QUERY
from enum import Enum
from typing import Any, List, Optional, Union
from arkitekt.contracts.reservation import Reservation
from arkitekt.models.graphql import GraphQLModel
from herre.loop import loopify, loopify_gen
from rich.table import Table




class Node(GraphQLModel):
    name: Optional[str]
    description: Optional[str]
    package: Optional[str]
    interface: Optional[str]
    args: Optional[List[Union[StructureArgPort, StringArgPort, IntArgPort, ListArgPort, DictArgPort]]]
    kwargs: Optional[List[Union[StructureKwargPort, StringKwargPort, IntKwargPort, ListKwargPort, DictKwargPort]]]
    returns: Optional[List[Union[StructureReturnPort, StringReturnPort, IntReturnPort, ListReturnPort, DictReturnPort]]]
    type: Optional[NodeType]


    def reserve(self, loop=None, monitor = None, ignore_node_exceptions=False, bounced=None, **params) -> Reservation:
        return Reservation(self, loop=loop, monitor=monitor, ignore_node_exceptions=ignore_node_exceptions, bounced=bounced, **params)


    async def call_async_func(self,*args,  reserve_params={}, **kwargs):
        async with self.reserve(**reserve_params) as res:
            return await res.assign_async(*args, **kwargs)


    async def call_async_gen(self,*args,  reserve_params={}, **kwargs):
        async with self.reserve(**reserve_params) as res:
            async for result in res.stream_async(*args, **kwargs):
                yield result


    def __call__(self, *args: Any, fill_graphical= True, **kwargs) -> Any:
        """Call

        Call is a convenience on max function, its reserves the Node and wraps it either as
        an geneator (both async and non async depending on context) or call it as a function
        this should only be done if you know what you are doing. 
        
        Args:
            reserve_params (dict, optional): [description]. Defaults to {}.

        Returns:
            Any: Generator or Function
        """
        if fill_graphical: args, kwargs = fill_args_kwargs_graphically(self, args, kwargs)

        if self.type == NodeType.FUNCTION:
            return loopify(self.call_async_func(*args, **kwargs))

        if self.type == NodeType.GENERATOR:
            return loopify_gen(self.call_async_gen(*args, **kwargs))


    class Meta:
        register = False
        ward = "arkitekt"
        get = NODE_GET_QUERY
        create = NODE_CREATE_QUERY

    def __rich_repr__(self):
        yield self.name
        yield "args", self.args
        yield "kwargs", self.kwargs
        yield "returns", self.returns

    def __rich__(self):
        my_table = Table(title=f"Node: {self.name}", show_header=False)

        my_table.add_row("ID", str(self.id))
        my_table.add_row("Package", self.package)
        my_table.add_row("Interface", self.interface)
        my_table.add_row("Type", self.type)

        return my_table


    def _repr_html_(self):
        return f"""
        <div class="container" style="border:1px solid #00000f;padding: 4px;">
            <div class="item item-1 font-xl">{self.name}</div>
            <div class="item item-2">{self.package}/{self.interface}</div>
            <div class="item item-3">Args: {" ".join([port._repr_html_list() for port in self.args])}</div>
            <div class="item item-3">Kwargs: {" ".join([port.key for port in self.kwargs])}</div>
            <div class="item item-3">Returns: {" ".join([port.key for port in self.returns])}</div>
        </div>"""








from herre.console.context import get_current_console
from typing import Any, Dict, List, Tuple
import asyncio


class SerializationError(Exception):
    """Serialization Error

    Serialization Errors are raised when a Shrinking or Expanding Error
    occurs during a shrinking/expanding event
    """
    pass



class ShrinkingError(SerializationError):
    """ Shrinking Error
    
    Is being rased from a Port Shrink Exception
    
    """

class ExpandingError(SerializationError):
    """ Expanding Error
    
    Is being rased from a Port Expand Exception
    
    """





async def shrink_inputs(node, *args, **kwargs) -> Tuple[List[Any], Dict[str, Any]]:
    """Shrinks args and kwargs 

    Shrinks the inputs according to the Node Definition

    Args:
        node (Node): The Node

    Raises:
        ShrinkingError: If args are not Shrinkable
        ShrinkingError: If kwargs are not Shrinkable

    Returns:
        Tuple[List[Any], Dict[str, Any]]: Parsed Args as a List, Parsed Kwargs as a dict
    """
    


    
    shrinked_args_futures = [port.shrink(arg) for port, arg in zip(node.args, args)]

    try:
        shrinked_args = await asyncio.gather(*shrinked_args_futures) # Worrysome because others won't be cancelled on Exception
    except Exception as e:
        raise ShrinkingError(f"Couldn't shrink Arguments {args}") from e

    try:
        shrinked_kwargs = {
            port.key: await port.shrink(kwargs.get(port.key, port.default))
            for port in node.kwargs
        }
    except Exception as e:
        raise ShrinkingError(f"Couldn't shrink KeywordArguments {kwargs}") from e

    return shrinked_args, shrinked_kwargs
    

async def expand_inputs(node, args: List[Any], kwargs: Dict[str, Any]) -> Tuple[List[Any], Dict[str, Any]]:
    """Expands Args and Kwargs

    Expands the Args and Kwargs according to the Node definition 

    Args:
        node (Node): Nodes
        args (List[Any]): Srhinked Args
        kwargs (Dict[str, Any]): Shrinkag Kwargs

    Raises:
        ExpandingError:  If args are not expandable
        ExpandingError:  If kwargs are not expandable

    Returns:
        Tuple[List[Any], Dict[str, Any]]: Expanded Args, Expanded Kwargs
    """

    expanded_args_futures = [port.expand(arg) for port, arg in zip(node.args, args)]
    try:
        expanded_args = await asyncio.gather(*expanded_args_futures)
    except Exception as e:
        raise ExpandingError(f"Couldn't expand Args {args}") from e

    try:
        expanded_kwargs = {
            port.key: await port.expand(kwargs.get(port.key, port.default))
            for port in node.kwargs
        }
    except Exception as e:
        raise ExpandingError(f"Couldn't expand Kwargs {kwargs}") from e


    return expanded_args, expanded_kwargs


async def shrink_outputs(node, returns) -> List[Any]:
    """Shrinks returns

    Shrinks the returns according to the Node Definition

    Args:
        node (Node): The Node
        returns (List[any]): The Returns

    Raises:
        ShrinkingError: If returns are not Shrinkable

    Returns:
        List[Any]: Parsed Returns
    """
    if not isinstance(returns, list) and not isinstance(returns, tuple): returns = [returns] # We are dealing with a single output, convert it to a proper port like structure
    shrinked_returns_future = [port.shrink(val) for port, val in zip(node.returns, returns)]
    try:
        return await asyncio.gather(*shrinked_returns_future)
    except Exception as e:
        raise ShrinkingError(f"Couldn't shrink Returns {returns}") from e


async def expand_outputs(node, returns) -> List[Any]:
    """Expands Returns

    Expands the Returns according to the Node definition 


    Args:
        node (Node): Node definition
        returns (List[any]): The returns

    Raises:
        ExpandingError: if they are not expandable

    Returns:
        List[Any]: The Expanded Returns
    """
    
    expanded_returns_future = [port.expand(val) for port, val in zip(node.returns, returns)]
    try:
        returns = await asyncio.gather(*expanded_returns_future)
        if len(returns) == 1: 
            return returns[0] # We are dealing with a single output, just cast back
        else:
            return returns
    except Exception as e:
        raise ExpandingError(f"Couldn't expand Returns {returns}") from e



from arkitekt.packers.models.base import StructureModel
from arkitekt.models.graphql import GraphQLModel


class GraphQLStructure(GraphQLModel, StructureModel):
    __typename: str

    @classmethod
    async def expand(cls, identifier):
        return await cls.asyncs.get(id=identifier)

    async def shrink(self):
        assert self.id is not None, "In order to send a Model through a Port you need to query 'id' in your GQL Query"
        return self.id


    class Meta:
        abstract = True